unit Perlin;

{Usage:
 First of all you should initialize all values (the standard values won't satisfy you :-) )
 For example:
  SetPersistence(0.6);
  SetSeed(12345);
  SetDepth(4);
 The amplitude depends on your application. For RGB values you usually require:
  SetAmplitude(0.0, 255.0);
 After you've called Create you can start producing values on demand. Make sure you
 call the GetValue function with an interval of about 0.01 depending on your application. If
 you use an interval of >= 1 you won't notice any smoothing!
}

interface

uses math, sysutils, dialogs;

type
  TPerlinArray = array of array of Single;

  TRandomMode = (rmPerlin, rmRandom);

  TPerlinNoise2D = class
  protected
    function Interpolate(V1, V2, Step: Single): Single;
    function InterpolatedRand(X, Y: Single): Single;
    function InterpolatedOctave(X, Y: Single; pOctave: integer): Single;
    function Rand(X, Y: Integer): Single;
  public
    Seed: integer;
    Depth: integer;
    Amplitude: Single;
    m_Add: Single;
    Persistence: Single;
    Frequency: Single;
    FrequencyFactor: Single;
    Size: integer;

    m_TmpSeed: Integer;

    Layer: Integer; // Welche Ebene soll dargestellt werden? Wenn -1, werden alle dargestellt

    PerlinArray: TPerlinArray;

    Octaves: array {Oktavenzahl} of TPerlinArray;

    // Boolean Fields
    TextureMode: boolean; // Kachelbar
    // End of Boolean Fields

    RandomMode: TRandomMode;

    procedure SetValues; virtual;

    // Set the minimal and maximal values for the result
    procedure SetAmplitude(pMin, pMax: Single);

    constructor Create;

    // Begriffsklrung
    {
      Absolut = Absolute Position in Pixeln. Bei einer Size von 512 => Wertebereich 0..511
      Relativ = Gleitkomma-Wert. Bereich 0..1
      Oktave  = Jede Oktave ist unterschiedlich gro (Doppelt so gro, wie Vorgnger)
                Ganzzahl. Bereich 0..Oktavengre

      AbsoluteOctave
    }

    procedure Calculate;
    function GetValueFromOctaveByAbsolute(X, Y: integer; pOctave: integer): Single; // Liefert den Wert result einer Oktave Octave an der Absoluten Position XY zurck.
    function GetValue(X, Y: Single): Single; overload; virtual;
    function GetValue(X, Y: Single; Layer: integer): Single; overload; virtual;
  end;

implementation

{ TPerlinNoise2D }

procedure TPerlinNoise2D.SetValues;
var
  L_Range: Single;
  L_SummedMultiplyer: Single;
  L_Multiplyer: Single;
  L_Count: Integer;
begin
  L_Range := Amplitude - m_Add;
  m_Add := m_Add + L_Range * 0.5;

  // We get values between -1 and 1 out of the Rand function
  L_SummedMultiplyer := 0;
  L_Multiplyer       := 2;

  for L_Count := 0 to Depth -1 do
  begin
    L_SummedMultiplyer := L_SummedMultiplyer + L_Multiplyer;
    L_Multiplyer := L_Multiplyer * Persistence;
  end;

  Amplitude := L_Range / L_SummedMultiplyer;
end;

procedure TPerlinNoise2D.SetAmplitude(pMin, pMax: Single);
begin
  m_Add     := pMin;
  Amplitude := pMax;
end;

constructor TPerlinNoise2D.Create;
begin
  inherited Create;
  Persistence := 0.6;
  Seed := 12345;
  Depth := 8;
  SetAmplitude(0, 255);
  Frequency := 0.01;
  FrequencyFactor := 2;
  Size := 256;
end;

function TPerlinNoise2D.GetValue(X, Y: Single): Single;
var
  L_Total: Single;
  L_Amplitude: Single;
  L_Count: Integer;
  L_Frequency: Single;
begin
  L_Total := 0;
  L_Amplitude := Amplitude;
  L_Frequency := Frequency;
  m_TmpSeed := Seed;

  for L_Count := 0 to Depth -1 do
  begin
    L_Total := L_Total + InterpolatedOctave(X, Y, L_Count) * L_Amplitude;
    L_Amplitude := L_Amplitude * Persistence; // L_Amplitude seems to be Faktor
    L_Frequency := L_Frequency * FrequencyFactor;
    m_TmpSeed := Round (m_TmpSeed * FrequencyFactor); // Diese Zeile entfernen, um fraktalen Charakter zu zeigen ;)
  end;

  L_Total := L_Total + m_Add;
  result := L_Total;
end;

function TPerlinNoise2D.GetValue(X, Y: Single; Layer: integer): Single; // Only returns a specific layer
var
  L_Total: Single;
  L_Amplitude: Single;
  L_Count: Integer;
  L_Frequency: Single;
begin
  L_Total := 0;
  L_Amplitude := Amplitude;
  L_Frequency := Frequency;
  m_TmpSeed := Seed;

  begin
    L_Total := L_Total + InterpolatedOctave(X, Y, Layer) * L_Amplitude;
    L_Amplitude := L_Amplitude * Persistence; // L_Amplitude seems to be Faktor
    L_Frequency := L_Frequency * FrequencyFactor;
    m_TmpSeed := Round (m_TmpSeed * FrequencyFactor); // Diese Zeile entfernen, um fraktalen Charakter zu zeigen ;)
  end;

  L_Total := L_Total + m_Add;
  result := L_Total;
end;

function TPerlinNoise2D.InterpolatedRand(X, Y: Single): Single;
var
  L_IntX: Integer;
  L_FracX: Single;

  L_IntY: Integer;
  L_FracY: Single;

  L_V1, L_V2, L_V3, L_V4, L_ResX1, L_ResX2: Single;
begin
  L_IntX  := Trunc(X);
  L_FracX := X - L_IntX;

  L_IntY  := Trunc(Y);
  L_FracY := Y - L_IntY;

  L_V1 := Rand(L_IntX,     L_IntY    );
  L_V2 := Rand(L_IntX + 1, L_IntY    );
  L_V3 := Rand(L_IntX,     L_IntY + 1);
  L_V4 := Rand(L_IntX + 1, L_IntY + 1);

  L_ResX1 := Interpolate(L_V1, L_V2, L_FracX);
  L_ResX2 := Interpolate(L_V3, L_V4, L_FracX);

  result := Interpolate(L_ResX1, L_ResX2, L_FracY);
end;

function TPerlinNoise2D.InterpolatedOctave(X, Y: Single; pOctave: integer): Single;
var
  L_IntX: Integer;
  L_FracX: Single;

  L_IntY: Integer;
  L_FracY: Single;

  L_V1, L_V2, L_V3, L_V4, L_ResX1, L_ResX2: Single;
begin
  L_IntX  := Trunc(X);
  L_FracX := X - L_IntX;

  L_IntY  := Trunc(Y);
  L_FracY := Y - L_IntY;

  L_V1 := GetValueFromOctaveByAbsolute(L_IntX,     L_IntY    , pOctave);
  L_V2 := GetValueFromOctaveByAbsolute(L_IntX + 1, L_IntY    , pOctave);
  L_V3 := GetValueFromOctaveByAbsolute(L_IntX,     L_IntY + 1, pOctave);
  L_V4 := GetValueFromOctaveByAbsolute(L_IntX + 1, L_IntY + 1, pOctave);

  L_ResX1 := Interpolate(L_V1, L_V2, L_FracX);
  L_ResX2 := Interpolate(L_V3, L_V4, L_FracX);

  result := Interpolate(L_ResX1, L_ResX2, L_FracY);
end;

function TPerlinNoise2D.Interpolate(V1, V2, Step: Single): Single;
var
  L_Fact: Single;
begin
  // This interpolation methode is faster than a cosinus function, but it's neary
  // equal to the cos-graph. I simply take a square function either in positiv
  // direction or in negativ direction starting from 1.0.

  if( Step > 0.5 ) then
  begin
    L_Fact := (1.0 - Step);
    L_Fact := L_Fact * L_Fact * 2.0;
  end
  else
  begin
    L_Fact := 1.0 - Step * Step * 2.0;
  end;

  result := V1 * L_Fact + V2 * (1.0 - L_Fact);
  //result := V1;
end;

function TPerlinNoise2D.Rand(X, Y: Integer): Single;
var
  L_Base: Integer;
begin
  case RandomMode of
    rmPerlin:
      begin
        L_Base := X + Y * 57 + m_TmpSeed;
        L_Base := (L_Base shl 13) xor L_Base;
        result := ( 1 - ( (L_Base * (L_Base * L_Base * 15731 + 789221) + 1376312589) and $7fffffff) * ( 1 / 1073741824 ));
      end;

    rmRandom:
      result := random;
  end;
end;

procedure TPerlinNoise2D.Calculate;
var
  L_Amplitude: Single;
  L_Frequency: Single;

  i: integer;
  X,Y: integer;
begin
  L_Amplitude := Amplitude;
  L_Frequency := Frequency;
  m_TmpSeed := Seed;
  setlength(Octaves, 0);
  setlength(Octaves, Depth);
  for i := 0 to Depth -1 do
  begin
    setlength(Octaves[i], Round(Power(2, i+1)) +1, Round(Power(2, i+1)) +1);
    for X := 0 to HIGH(Octaves[i]) -1 do
    begin
      for Y := 0 to HIGH(Octaves[i][X]) -1 do
      begin
        Octaves[i][X,Y] := Rand(Round(X * L_Frequency), Round(Y * L_Frequency)) * L_Amplitude;
        //Octaves[i][X,Y] := i * 1;
        //if TextureMode then CopyBorders;
      end;
    end;
    L_Amplitude := L_Amplitude * Persistence; // L_Amplitude seems to be Faktor
    L_Frequency := L_Frequency * FrequencyFactor;
    m_TmpSeed := Round (m_TmpSeed * FrequencyFactor);
  end;
end;

function TPerlinNoise2D.GetValueFromOctaveByAbsolute(X, Y: integer; pOctave: integer): Single;
var
  OctaveSize: integer;
  ArrayX,
  ArrayY: integer;
begin
  OctaveSize := Round(Power(2, pOctave + 1));
  ArrayX := Trunc(X / Size * OctaveSize);
  ArrayY := Trunc(Y / Size * OctaveSize);
  if ArrayX > OctaveSize then Abort;
  if ArrayY > OctaveSize then showmessage('Y: ' + inttostr(Y) + ' ArrayY: ' + inttostr(ArrayY) + ' pOctave: ' + inttostr(pOctave) + ' OctaveSize: ' + inttostr(OctaveSize));
  //if ArrayX < 0 then halt;
  //if ArrayY < 0 then halt;
  result := Octaves[pOctave][ArrayX][ArrayY];
end;

end.
